// Zilch:  A game strategy optimizer for Zilch, Farkle and similar dice games.
// Copyright 2010 Matthew T. Busche.
//
// This program is free software:  you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// This program is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
// more details.
//
// You should have received a copy of the GNU General Public License along
// with this program.  If not, see <http://www.gnu.org/licenses/>.

import java.util.TreeSet;

/** A game strategy optimizer for Zilch, Farkle and similar dice games.
 ** Please read the accompanying document at
 ** http://www.mattbusche.org/article/zilch/ for a better understanding
 ** of the workings of the software.
 **/
public class Zilch {

    // Normally I use rather verbose variable naming, but for this
    // software I follow the naming in the associated paper that derives
    // the technique used to solve for the optimal zilch playing implemented
    // in this software strategy.  Since that paper is mathematical in form,
    // the variable names are quite short (at least for those concepts that
    // map directly to the paper).

    // The variables NUM_DICE and NUM_PIPS give the number of dice used in
    // the game and the number of pips on a die, respectively.  Changing
    // these values will not produce any kind of meaningful results.  The
    // game rules are setup specifically for six six sided dice, and there
    // are a few places where bits of code depend on these variables having a
    // value of 6.  These variables are introduced primarily just to make it
    // clear where I'm iterating over the number of dice used in the game vs
    // the number of pips on a die.
    //
    static final int NUM_DICE  = 6;
    static final int NUM_PIPS  = 6;

    // You can change these scoring constants to match the rules of the game
    // you play.  All scores must be multiples of 50.  Setting a score to
    // zero effectively removes it as a scoring option.  So for farkle,
    // where a NOTHING roll is a 6-die farkle (as opposed to 500 points)
    // just set NOTHING_SCORE to zero.
    //
    static final int FIVE_SCORE       = 50;
    static final int ONE_SCORE        = 100;
    static final int THREE_PAIR_SCORE = 1500;
    static final int STRAIGHT_SCORE   = 1500;
    static final int NOTHING_SCORE    = 500;
    static final int SET_SCORE[][]    = {
        {1000, 2000, 4000, 8000},
        { 200,  400,  800, 1600},
        { 300,  600, 1200, 2400},
        { 400,  800, 1600, 3200},
        { 500, 1000, 2000, 4000},
        { 600, 1200, 2400, 4800}
    };

    /** The fewest number of points you can bank.  This varies from
     ** game variant to game variant as well.  Just pick the value
     ** that applies to you.
     **/
    static final int SMIN = 300;

    /** If it is possible to zilch (or farkle or whatever you call a
     ** non-scoring roll) with 6 dice, then the optimal strategy for scores
     ** above SBIG is simply to always stop rolling and bank.  If it is NOT
     ** possible to zilch with 6 dice, then the optimal strategy for scores
     ** above SBIG is to bank any time you are faced with rolling less than 6
     ** dice, and to always roll when you have 6 dice to throw.  The simplest
     ** correct thing to do would be to simply set SBIG to some humongous
     ** value and forget about it, but run times are linear in SBIG, so
     ** instead I've written the initSBIG routine finds the value of SBIG that
     ** satisfies these properties.
     **/
    static int SBIG = 0;

    /** numWaysToRollNDice[n] is just the number of ways to roll n dice or
     ** just 6^n.
     **/
    static final int[] numWaysToRollNDice = new int[NUM_DICE+1];

    static
    {
        numWaysToRollNDice[0] = 1;
        for(int n = 1; n <= NUM_DICE; ++n)
            numWaysToRollNDice[n] = numWaysToRollNDice[n-1] * NUM_PIPS;
    }

    /** If you already have SBIG or more points accumlated on the current turn
     ** and have a 6 die roll coming up; then EBIG6 is the expected increase
     ** in your score if you play in such a way to maximize your final
     ** expected score.  If you have a non-zero zilch probability when throwing
     ** 6 dice, then EBIG6 is simply zero, but otherwise EBIG6 is a non-zero
     ** constant.  The routine initEBIG6 finds this value.
     **/
    static double EBIG6;

    /** Additional points lost in the event of a zilch.  You can set this
     ** to determine the best way to play when you already have two
     ** consecutive zilches and will lose additional points if you zilch
     ** again.  This is also used to introduce an artifical zilch penalty to
     ** reduce zilch probability at the cost of lowered turn score
     ** expectations, which ultimately allows optimization of the long-term
     ** expected point gain across all turns when considering the affects of
     ** three consecutive zilch rule.
     **/
    static double y = 0.0;

    /** E[s/50][n-1] gives the number of additional points you can expect to
     ** make for the remainder of the turn if you roll the dice when you
     ** currently have accumlated s points (for the turn), have n dice
     ** available for rolling, and play optimally for the remainder of the
     ** turn.  If E[s/50][n-1] is less than 0, then you should bank rather
     ** than roll.
     **
     ** Consider this example:  suppose on your opening roll you roll three
     ** threes and a single five.  You have these four choices:
     **
     ** (1) take the single 5 for 50 points and roll 5 dice,
     ** (2) take the three-3s for 300 points and roll 3 dice,
     ** (3) take the three-3s AND the five for 350 points and roll 2 dice, or
     ** (4) take the three-3s AND the five for 350 points and bank.
     **
     ** For two of these three cases you also have the decision of whether to
     ** roll or bank.  To figure out what you should do, compare these value:
     **
     ** (1) E[50/50][5-1] + 50   // take 50 points and roll 5 dice;
     ** (2) E[300/50][3-1] + 300 // take 300 points and roll 3 dice;
     ** (3) E[350/50][2-1] + 350 // take 350 points and roll 2 dice; or
     ** (4) 350                  // take 350 points and bank.
     **
     ** Your expected score is maximized by choosing the case with the
     ** greatest value.
     **/
    static double E[][];

    /** p[s/50][n-1] gives the probability of entering state (s, n)
     ** during the turn.  I calculate these only to come up with
     ** the probability of zilching for the whole turn, z.
     **/
    static double psn[][];

    /** z gives the probability of zilching on a turn.
     **/
    static double z;

    /** This structure is used to store a possible score from a roll
     ** and also encodes the number of dice used in the roll.  Note
     ** that the default value of s=0, n=0 corresponds to a zilch.
     **/
    static class Score {
        int s;  // score
        int n;  // num dice used

        Score set(Score c) {
            this.s = c.s;
            this.n = c.n;
            return this;
        }

        public Score set(int s, int n) {
            // As a convenience, if the caller passes in an score value of 0
            // then force n to zero as well.  This is useful for when
            // users change a scoring value of some roll to zero (i.e.,
            // removes that type of score from the game.)  This happens
            // with the NOTHING roll between different game variants.
            // I've also seen game variants where 4 of a kind and 5 of
            // a kind is worth nothing.  (I.e., you can only score 3 of
            // a kind.)
            //
            this.s = s;
            this.n = (s == 0) ? 0 : n;
            return this;
        }
    }

    static class RollStats {

        /** Number of times a roll resulted in a zilch.
         **/
        int numZilchRoll;

        /** Number of times a roll resulted in a subsequent free roll.
         **/
        int numFreeRoll;

        /** Number of times a roll resulted in neither a zilch nor a free roll.
         **/
        int numOtherRoll;

        /** Total number of points for all rolls that resulted in
         ** subsequent free roll.
         **/
        int freeRollSum;

        /** Total number of points for all rolls that did NOT result in
         ** subsequent free roll.
         **/
        int otherRollSum;

        /** The total of all the expected score changes for all rolls given the
         ** initial state (s, n).  Must divide by numWaysToRollNDice[n] to get the actual
         ** expected score change for a roll made while in (s, n).
         **/ 
        double expectedChangeSum;
    }

    public static void main(String[] args) {
        try {
            if(args.length > 0)
                y = Double.parseDouble(args[0]);

            init();
            calcE();
            calcPsnAndZ();
            showSummary();
            showE();

            // You don't need to know the  probability of entering every
            // possible state to play the strategy, but it's sorta cool info
            // if you want to print it out.
            //
            //showPsn();
        }
        catch(Exception x) {
            x.printStackTrace();
        }
    }

    /** Find expected change in score from each possible state.
     **/
    static void calcE() {
        for(int s = SBIG - 50; s >= 0; s -= 50)
            for(int n = NUM_DICE; n >= 1; --n)
                E[s/50][n-1] = calcE(s, n);
    }

    /** Calculate E for the case of having s points currently accumlated in
     ** the turn and n dice available to roll.
     **/
    static double calcE(int s, int n) {
        int[] dice = new int[n];
        RollStats t = new RollStats();
        tryAllRolls(t, s, dice, 0);
        return t.expectedChangeSum / numWaysToRollNDice[n];
    }

    /** This function is used to reset the state variable n back to 6 when you
     ** select a score that uses all remaining dice, giving you a free roll.
     **/
    static int F(int x) {
        return x == 0 ? NUM_DICE : x;
    }

    // Examines all possible rolls of (dice.length) dice.  For each roll
    // permutation, the scoring combination that maximizes your final expected
    // score for the turn is selected.  Various useful statistics are
    // maintained in t which is assumed to be clear when the first call is
    // made.  The argument d is the recursion variable and should always
    // be set to 0 for the first call.
    //
    static void tryAllRolls(RollStats t, int s, int[] dice, int d) {
        int n = dice.length;
        if(d >= n) {
            Score c = C(dice, s);

            if(c.s == 0) {
                t.numZilchRoll++;
                t.expectedChangeSum -= (s+y);
            }
            else {
                int si = s + c.s;
                int ni = F(n - c.n);
                t.expectedChangeSum += (T(si, ni) - s);
                if(ni == NUM_DICE) {
                    t.numFreeRoll++;
                    t.freeRollSum += c.s;
                }
                else {
                    t.numOtherRoll++;
                    t.otherRollSum += c.s;
                }
            }
            return;
        }

        for(int i = 1; i <= NUM_PIPS; ++i) {
            dice[d] = i;
            tryAllRolls(t, s, dice, d+1);
        }
    }

    // Examines all possible scoring dice selections that could be taken for
    // the specific roll defined in the dice array and for each such scoring
    // obtains the expected score for the turn state entered when the score is
    // applied to the previous turn state.  C returns the Score that produces
    // the maximum expected score over these different scoring possiblities.
    // (I.e., it returns the scoring combination associated with the "best"
    // turn state that can be entered given the dice rolled.)
    //
    static Score C(int[] dice, int s) {
        final int n = dice.length;
        int[] bin = new int[NUM_PIPS+1];
        for(int i = 0; i < n; ++i)
            ++bin[dice[i]];

        Score best = new Score();
        Score work = new Score();

        tryAllSets(best, bin, s, n, 0, 0, 1);

        if(n == NUM_DICE) {
            // RUN OF 6
            //
            if(bin[1] == 1 && bin[2] == 1 && bin[3] == 1 && bin[4] == 1 && bin[5] == 1 && bin[6] == 1)
                updateBestScore(best, work.set(STRAIGHT_SCORE, 6), s, n);

            // THREE PAIR
            //
            else if(bin[1]%2==0 && bin[2]%2==0 && bin[3]%2==0 && bin[4]%2==0 && bin[5]%2==0 && bin[6]%2==0)
                updateBestScore(best, work.set(THREE_PAIR_SCORE, 6), s, n);

            // NOTHING: if no previous scoring option works, we have a NOTHING roll
            //
            if(best.s == 0)
                updateBestScore(best, work.set(NOTHING_SCORE, 6), s, n);
        }

        return best;
    }

    static void tryAllSets(Score best, int[] bin, int s, int n, int rollPoints, int rollUsed, int i) {

        // We've finished examining all set possibilities from sets of ones
        // through sets of sixes (branching on whether to take that set in
        // each case), now we take the remaining dice and examine all
        // combinations of scoring for any ones or fives that are remaining in
        // the roll.
        //
        if(i > NUM_PIPS) {
            tryAllOnes(best, bin, s, n, rollPoints, rollUsed);
            return;
        }

        // case 1: try not taking a set
        //
        tryAllSets(best, bin, s, n, rollPoints, rollUsed, i+1);

        // case 2: try taking a set (if it exists)
        //
        int binSize = bin[i];
        if(binSize >= 3) {

            // Need to temporarily set binSize to zero so ensure that
            // sets of 1s and 5s don't get scored again as single 1s and 5s.
            //
            bin[i] = 0;
            tryAllSets(best, bin, s, n, rollPoints + SET_SCORE[i-1][binSize-3], rollUsed + binSize, i+1);

            // Restore the binSize
            //
            bin[i] = binSize;
        }
    }

    static void tryAllOnes(Score best, int[] bin, int s, int n, int rollPoints, int rollUsed) {
        tryAllFives(best, bin, s, n, rollPoints, rollUsed);
        int max = Math.min(2, bin[1]);
        for(int i = 1; i <= max; ++i)
            tryAllFives(best, bin, s, n, rollPoints+i*ONE_SCORE, rollUsed+i);
    }

    static void tryAllFives(Score best, int[] bin, int s, int n, int rollPoints, int rollUsed) {
        Score work = new Score();
        updateBestScore(best, work.set(rollPoints, rollUsed), s, n);
        int max = Math.min(2, bin[5]);
        for(int i = 1; i <= max; ++i)
            updateBestScore(best, work.set(rollPoints+i*FIVE_SCORE, rollUsed+i), s, n);
    }

    /** Copies scoring combination c to best if c puts you into a turn state
     ** with a better final expected score for the turn.
     **/
    static void updateBestScore(Score best, Score c, int s, int n) {

        // The expression comparing the two T values below is only valid
        // if both c and best are non-zilching rolls.  If c is a zilch
        // then we just return; otherwise if best is a zilch then we
        // always set best to c.
        //
        if(c.s == 0)
            return;

        if(best.s == 0 || T(s + c.s, F(n - c.n)) > T(s + best.s, F(n - best.n)))
            best.set(c);
    }

    /** Returns the final expected score for a turn given that you have s
     ** points currently accumulated and have n dice left to roll.
     **/
    static double T(int s, int n) {
        if(s >= SBIG) {
            if(n == NUM_DICE)
                return s + EBIG6;
            else
                return s;
        }
        if(s >= SMIN)
            return s + Math.max(0.0, E[s/50][n-1]);
        else
            return s + E[s/50][n-1];
    }

    /** Calculates the turn zilch probability, z, and also calculates (and
     ** stores in the array psn) the probability of entering state (s, n) for
     ** all states (s, n) with s < SBIG.  Starting with the initial state
     ** (0,6), calcPsnAndZ walks back up the state tree enumerating all
     ** possible rolls from every possible source state and adding the
     ** probability of that roll times the probability of ever entering the
     ** source state to the probability of entering the destination state.
     ** Note that if you are supposed to bank from a state, then there are no
     ** possible rolls from that state.  Each time a zilch roll is thrown,
     ** probability is added to the zilch state as well.  Once we get to
     ** s == SBIG we can stop since there is no way to zilch once you reach
     ** that score.
     **/
    static void calcPsnAndZ()
    {
        psn[0][NUM_DICE-1] = 1;
        for(int s = 0; s < SBIG; s += 50)
            for(int n = NUM_DICE; n >= 1; --n)
                probabilityUpdate(s, n);
    }

    /** probabilityUpdate is invoked only when the probability of entering
     ** state (s, n) is already known.  This method immediately returns if
     ** you're not supposed to roll from this state (i.e., if you're supposed
     ** to bank making this a terminal state for the turn).  Otherwise it
     ** considers every possible roll out of this state and adds the
     ** probability of that roll (weighted by the probability of entering this
     ** state itself) to the state entered given that roll.  If the roll is a
     ** zilch the probability gets added to the zilch probability for the turn
     ** instead.  By walking up the table with increasing values of s, the
     ** probabilities of entering all possible states are calculable.  We stop
     ** once we get to s = SBIG since it's no longer possible to zilch once we
     ** reach that score.
     **/
    static void probabilityUpdate(int s, int n) {
        if(s >= SMIN && E[s/50][n-1] <= 0)
            return;
        int[] dice = new int[n];
        probabilityUpdate(s, dice, 0);
    }

    static void probabilityUpdate(int s, int[] dice, int d) {
        int n = dice.length;
        if(d >= n) {
            Score c = C(dice, s);
            if(c.s == 0)
                z += psn[s/50][n-1] / numWaysToRollNDice[n];
            else {
                int si = s + c.s;
                int ni = F(n - c.n);
                if(si < SBIG)
                    psn[si/50][ni-1] += psn[s/50][n-1] / numWaysToRollNDice[n];
            }
            return;
        }

        for(int i = 1; i <= NUM_PIPS; ++i) {
            dice[d] = i;
            probabilityUpdate(s, dice, d+1);
        }
    }

    static void init() {
        initEBIG6();     // You need to know EBIG6 before...
        initSBIG();      // ...you can find SBIG, and you need to know SBIG...
        initMatricies(); // ...to size the matricies properly.
    }

    static double BREAKPOINT_MAX;

    static void initEBIG6() {

        // If there's a non-zero probability of zilching with a 6 die roll
        // then EBIG6 will be 0.  For example, in Farkle you can throw a 6-die
        // farkle (same thing as a zilch), but in Zilch a 6 die roll that has
        // no normal scoring combination is actually worth 500 points.
        //
        // So the first thing I do is calculate the probability of zilching
        // with 6 dice, and if that is greater than zero, then I set EBIG6
        // to zero and I'm done!
        //
        EBIG6 = 0;
        int[] dice = new int[NUM_DICE];
        RollStats rs = new RollStats();
        tryAllRolls(rs, 0, dice, 0);
        if(rs.numZilchRoll != 0)
            return;

        // QUAD ONE PLUS PAIR PROBLEM:
        //
        // There are these curious rolls all involving exactly 4 ones:
        //
        //     1, 1, 1, 1, 2, 2
        //     1, 1, 1, 1, 3, 3
        //     1, 1, 1, 1, 4, 4
        //     1, 1, 1, 1, 6, 6
        //
        // These rolls can either be counted as four-ones for 2000 or 3-pair
        // for 1500.  Counting as 1500 uses all dice and offers a free roll.
        // Counting as 2000 only uses 4 dice (which pretty much forces you to
        // bank).  This leads to the interesting question of whether it is
        // better to take the 2000 points and bank or to take 1500 points and
        // get a free roll.  To answer the question, you need to know how much
        // a free roll is worth.  If just before you make one of these rolls
        // your current score is S, then if E(S+1500,6) is more than 500, then
        // you should take the 1500 points plus the free roll.  But if
        // E(S+1500,6) is less than 500 then you should take the 2000 points
        // and bank.
        //
        // This presents us with a bit of a conundrum since when this method
        // is called, no value for E is known.  Indeed, the purpose of this
        // method is to calculate and initialize our first E value, EBIG6,
        // which is the value of E for a 6 die roll when S is really big.
        // Once we know this, all other E values become calculable.
        //
        // EBIG6 is calculated using the tryAllRolls method passing in s=SBIG.
        // Because of the four rolls above, the behavior of tryAllRolls for
        // large s will depend upon the value of EBIG6 itself.  Specifically,
        // if EBIG6 has a value less than 500 then tryAllRolls will select
        // the four 1s as the best scoring combination; but if EBIG6 has a
        // value greater than 500, tryAllRolls will select three pair as the
        // best scoring combination.
        //
        // For normal zilch scoring rules, these are the only rolls in the
        // game where you can possibly score more points by NOT taking the
        // scoring combination that offers a free-roll, but it is possible to
        // change the scoring constants in such a way that there are other
        // similar scoring conflicts where one scoring choice gets you more
        // points but no free roll.
        //
        // The solution to this dilema is to first identify all of these
        // scoring conflicts for the given scoring constants.  For each
        // roll that has a scoring conflict, I store the difference between
        // the two ways to score that roll in a TreeSet named breakPoints.
        // When EBIG6 crosses over this breakpoint value, tryAllRolls will
        // switch it's choice about which scoring option to take.  Once all
        // the breakpoints are discovered, I iteratively seed EBIG6 with
        // values on each side of each breakpoint, invoke tryAllRolls for each
        // case and calculate the value of EBIG6 resulting from that seed
        // value.  The true value of EBIG6 will be the largest of these
        // calculated values.

        TreeSet<Double> breakPoints = new TreeSet<Double>();

        // I need a high-end break point that is guaranteed bigger than the
        // points from any scoring combination that uses less than 6 dice.
        //
        BREAKPOINT_MAX = 0;
        int u = Math.max(FIVE_SCORE, ONE_SCORE);
        BREAKPOINT_MAX = Math.max(BREAKPOINT_MAX, u*(NUM_DICE-1));
        for(int i = 0; i < NUM_PIPS; ++i)
            for(int j = 3; j < NUM_DICE; ++j)
                BREAKPOINT_MAX = Math.max(BREAKPOINT_MAX, SET_SCORE[i][j-3] + u*(NUM_DICE-j-1));
        ++BREAKPOINT_MAX;
        breakPoints.add(BREAKPOINT_MAX);

        findScoringConflicts(breakPoints, dice, 0);

        // Seed EBIG6 with a value midway between each neighboring pair
        // of scoring breakpoints, calculate EBIG6 resulting from each
        // seed, and pick the best one.
        //
        double bestEbig6 = 0;
        double lastBp = 0;
        for(double bp : breakPoints) {
            EBIG6 = (lastBp + bp) / 2;

            RollStats t = new RollStats();
            tryAllRolls(t, SBIG, dice, 0);

            // Probability of a 6 die roll resulting in a free roll.
            //
            double pF = (double) t.numFreeRoll / numWaysToRollNDice[NUM_DICE];

            // Expected number of points from a single roll of six dice
            // GIVEN that the 6 die roll results in a free roll.
            //
            double EF = (double) t.freeRollSum / t.numFreeRoll;

            // Expected number of points from a single roll of six dice
            // GIVEN that the 6 die roll does NOT result in a free roll.
            //
            double EB = (double) t.otherRollSum / t.numOtherRoll;

            double ebig6 = EB + pF/(1-pF) * EF;

            bestEbig6 = Math.max(bestEbig6, ebig6);

            lastBp = bp;
        }
        EBIG6=bestEbig6;
    }

    static void findScoringConflicts(TreeSet<Double> breakPoints, int[] dice, int d) {
        int n = dice.length;
        if(d >= n) {
            EBIG6=0;
            Score c1 = C(dice, SBIG);
            EBIG6=BREAKPOINT_MAX;
            Score c2 = C(dice, SBIG);
            if(c1.n != c2.n)
                breakPoints.add((double)(c1.s - c2.s));
            return;
        }

        for(int i = 1; i <= NUM_PIPS; ++i) {
            dice[d] = i;
            findScoringConflicts(breakPoints, dice, d+1);
        }
    }

    /** This method determines the smallest valid setting for SBIG and then
     ** sets SBIG to that valule.  If all we wanted was correctness, then we
     ** could just set SBIG to a very large number and forget about it.  The
     ** problem is the run times are linear in SBIG and the times are fairly
     ** long.  I'm not terribly concerned about performance for this
     ** appliation, but I really didn't like having SBIG hard-coded anyway
     ** since if it is ever set too small the program will output meaningless
     ** results.  So to eliminate the hard-coded constant and to speed up the
     ** program, I added this routine to do a binary search for the best value
     ** for SBIG.
     **/
    static void initSBIG() {
        int low = 0;
        int high = 2000;
        while(!isValidSBIG(high)) {
            low = high;
            high *= 2;
        }

        while(high - low > 50) {
            int mid = ((high + low) / 100) * 50;
            if(isValidSBIG(mid))
                high = mid;
            else
                low = mid;
        }
        SBIG = Math.max(high, SMIN + 50);
    }

    /** Returns true if the supplied score value is big enough to be a valid
     ** setting for SBIG.  There are two cases.  If EBIG6 is zero, find the
     ** smallest SBIG where calcE(SBIG-50,6) < 0.  Otherwise, find the
     ** smallest SBIG where calcE(SBIG-50,5) = EBIG6
     **/
    static boolean isValidSBIG(int s) {
        if(EBIG6 == 0)
            return calcE(s - 50, 6) < 0;
        else
            return calcE(s - 50, 5) < 0;
    }

    static void initMatricies() {
        E   = new double [SBIG/50][NUM_DICE];
        psn = new double [SBIG/50][NUM_DICE];
    }

    static void showSummary() {
        System.out.printf("Zilch penalty=%.3f\n", y);
        System.out.printf("Probability of zilching using this strategy=%8.6f\n", z);
        System.out.printf("Expected points for the turn=%10.6f\n", E[0][NUM_DICE-1]);
        System.out.printf("Expected points for the turn corrected for false zilch penalty=%10.6f\n", E[0][NUM_DICE-1] + z*y);
    }

    static void showE() {
        System.out.printf("\nExpected change in score if you roll n dice while having s points accumlated:\n");
        System.out.printf("   s | n");
        for(int n = NUM_DICE; n >= 1; --n)
            System.out.printf("     %d     ", n);
        System.out.printf("\n");
        for(int s = SBIG - 50; s >= 0; s -= 50) {
            System.out.printf("%5d:", s);
            for(int n = NUM_DICE; n >= 1; --n)
                if(psn[s/50][n-1] > 0)
                    System.out.printf(" %10.3f", E[s/50][n-1]);
                else
                    System.out.printf("          *");
            System.out.printf("\n");
        }
    }

    static void showPsn() {
        System.out.printf("\nProbability of entering a state with s points accumlated and n dice to roll:\n");
        System.out.printf("   s | n");
        for(int n = NUM_DICE; n >= 1; --n)
            System.out.printf("   %d     ", n);
        System.out.printf("\n");
        for(int s = SBIG - 50; s >= 0; s -= 50) {
            System.out.printf("%5d:", s);
            for(int n = NUM_DICE; n >= 1; --n)
                if(psn[s/50][n-1] > 0)
                    System.out.printf(" %8.6f", psn[s/50][n-1]);
                else
                    System.out.printf("        *");
            System.out.printf("\n");
        }
    }
}
